part of masamune_purchase_stripe;

final stripeConnectAccountModelProvider = ChangeNotifierProvider(
  (_) => StripeConnectAccountModel(),
);

class StripeConnectAccountModel
    extends ValueModel<FirestoreDynamicDocumentModel?> {
  StripeConnectAccountModel() : super(null);

  static const String _callbackUrl = "https://stripe.apps.mathru.net";

  @override
  @protected
  void initState() {
    super.initState();
    _initialize();
  }

  /// Returns itself after the load finishes.
  Future<void>? get loading => _loadingCompleter?.future;
  Completer<void>? _loadingCompleter;

  /// Returns itself after the save finishes.
  Future<void>? get saving =>
      _savingCompleter?.future ?? _registrationCompleter?.future;
  Completer<CommercePlatformStatus>? _registrationCompleter;
  Completer<void>? _savingCompleter;

  /// If this value is `true`,
  /// the change will be notified when [value] itself is changed.
  @override
  bool get notifyOnChangeValue => false;

  Future<void> _initialize() async {
    if (value != null) {
      return;
    }
    if (!StripeConnectCore.isInitialized) {
      await StripeConnectCore._future;
      if (!StripeConnectCore.isInitialized) {
        throw Exception("StripeConnectCore has not been initialized.");
      }
    }
    value = StripeConnectCore._accountModel;
    value?.addListener(_handledOnUpdate);
    notifyListeners();
  }

  void _handledOnUpdate() {
    notifyListeners();
  }

  bool get exists {
    if (value == null) {
      return false;
    }
    return value.containsKey(StripeConnectCore._accountKey);
  }

  bool get registered {
    if (value == null) {
      return false;
    }
    if (!value.containsKey(StripeConnectCore._accountKey)) {
      return false;
    }
    return value.containsKey(StripeConnectCore._capabilityKey) &&
        value
            .getAsMap(StripeConnectCore._capabilityKey)
            .containsKey("transfers");
  }

  Future<void> delete() async {
    if (_savingCompleter != null) {
      return saving;
    }
    _savingCompleter = Completer<void>();
    try {
      await _initialize();
      final uid = StripeConnectCore._debugUserId ?? userId;
      if (uid.isEmpty) {
        throw Exception("You are not logged in. Please log in once.");
      }
      if (!value.containsKey(StripeConnectCore._accountKey)) {
        throw Exception(
            "Account information is empty. Please run [create] method.");
      }
      final functions =
          read(functionsDocumentProvider(StripeConnectCore._serverPath));
      final res = await functions.call(
        parameters: {
          "mode": _StripeConnectAPIMode.deleteAccount.id,
          "userId": uid,
        },
      );
      if (res.isEmpty) {
        throw Exception("Response is invalid.");
      }
      notifyListeners();
      _savingCompleter?.complete();
      _savingCompleter = null;
    } catch (e) {
      _savingCompleter?.completeError(e);
      _savingCompleter = null;
      rethrow;
    } finally {
      _savingCompleter?.complete();
      _savingCompleter = null;
    }
  }

  Future<void> dashboard(
    BuildContext context, {
    required void Function(
      Widget webView,
      VoidCallback onClosed,
    )
        builder,
    bool popWhenClosedByWebview = false,
  }) async {
    if (_loadingCompleter != null) {
      return saving;
    }
    _loadingCompleter = Completer<void>();
    try {
      await _initialize();
      final uid = StripeConnectCore._debugUserId ?? userId;
      if (uid.isEmpty) {
        throw Exception("You are not logged in. Please log in once.");
      }
      if (!value.containsKey(StripeConnectCore._accountKey)) {
        throw Exception(
            "Account information is empty. Please run [create] method.");
      }
      if (!value.containsKey(StripeConnectCore._capabilityKey) ||
          !value
              .getAsMap(StripeConnectCore._capabilityKey)
              .containsKey("transfers")) {
        throw Exception("Your account has not been registered yet.");
      }
      final functions =
          read(functionsDocumentProvider(StripeConnectCore._serverPath));
      final res = await functions.call(
        parameters: {
          "mode": _StripeConnectAPIMode.dashboard.id,
          "userId": uid,
          "returnUrl": "$_callbackUrl/dashboard/finished",
        },
      );
      if (res.isEmpty) {
        throw Exception("Response is invalid.");
      }
      final endpoint = res.get("endpoint", "");
      if (endpoint.isEmpty) {
        throw Exception("Response is invalid.");
      }
      final onClosed = () {
        _loadingCompleter?.complete();
        _loadingCompleter = null;
      };
      final webView = _StripeConnectWebview(Uri.parse(endpoint),
          shouldOverrideUrlLoading: (controller, url) {
        final path = url.trimQuery().replaceAll(_callbackUrl, "");
        switch (path) {
          case "/dashboard/finished":
          case "https://connect.stripe.com/express_login":
            if (popWhenClosedByWebview) {
              context.navigator.pop();
            }
            onClosed.call();
            return NavigationActionPolicy.CANCEL;
        }
        return NavigationActionPolicy.ALLOW;
      }, onCloseWindow: (controller) {
        onClosed.call();
      });
      builder.call(webView, onClosed);
      await _loadingCompleter!.future;
      notifyListeners();
      _loadingCompleter?.complete();
      _loadingCompleter = null;
    } catch (e) {
      _loadingCompleter?.completeError(e);
      _loadingCompleter = null;
      rethrow;
    } finally {
      _loadingCompleter?.complete();
      _loadingCompleter = null;
    }
  }

  Future<CommercePlatformStatus> create(
    BuildContext context, {
    required void Function(
      Widget webView,
      VoidCallback onSuccess,
      VoidCallback onCancel,
    )
        builder,
    bool popWhenClosedByWebview = false,
  }) async {
    if (_registrationCompleter != null) {
      return _registrationCompleter!.future;
    }
    _registrationCompleter = Completer<CommercePlatformStatus>();
    try {
      await _initialize();
      final uid = StripeConnectCore._debugUserId ?? userId;
      if (uid.isEmpty) {
        throw Exception("You are not logged in. Please log in once.");
      }
      final functions =
          read(functionsDocumentProvider(StripeConnectCore._serverPath));
      final res = await functions.call(
        parameters: {
          "mode": _StripeConnectAPIMode.createAccount.id,
          "userId": uid,
          "refreshUrl": "$_callbackUrl/create_account/refresh",
          "returnUrl": "$_callbackUrl/create_account/success",
        },
      );
      if (res.isEmpty) {
        throw Exception("Response is invalid.");
      }
      if (res.get("next", "") == "registration") {
        final endpoint = res.get("endpoint", "");
        if (endpoint.isEmpty) {
          throw Exception("Response is invalid.");
        }
        final onSuccess = () {
          _registrationCompleter?.complete(CommercePlatformStatus.success);
          _registrationCompleter = null;
        };
        final onCancel = () {
          _registrationCompleter?.complete(CommercePlatformStatus.canceled);
          _registrationCompleter = null;
        };
        final webView = _StripeConnectWebview(Uri.parse(endpoint),
            shouldOverrideUrlLoading: (controller, url) {
          final path = url.trimQuery().replaceAll(_callbackUrl, "");
          switch (path) {
            case "/create_account/success":
              if (popWhenClosedByWebview) {
                context.navigator.pop();
              }
              onSuccess.call();
              return NavigationActionPolicy.CANCEL;
            case "/create_account/refresh":
              if (popWhenClosedByWebview) {
                context.navigator.pop();
              }
              onCancel.call();
              return NavigationActionPolicy.CANCEL;
          }
          return NavigationActionPolicy.ALLOW;
        }, onCloseWindow: (controller) {
          onCancel.call();
        });
        builder.call(webView, onSuccess, onCancel);
        final status = await _registrationCompleter!.future;
        notifyListeners();
        return status;
      } else {
        _registrationCompleter
            ?.complete(CommercePlatformStatus.alreadyRegistered);
        _registrationCompleter = null;
        notifyListeners();
        return CommercePlatformStatus.alreadyRegistered;
      }
    } catch (e) {
      _registrationCompleter?.completeError(e);
      _registrationCompleter = null;
      rethrow;
    } finally {
      _registrationCompleter?.complete(CommercePlatformStatus.canceled);
      _registrationCompleter = null;
    }
  }

  @override
  void dispose() {
    super.dispose();
    value?.removeListener(_handledOnUpdate);
  }
}
